<?php
namespace Zeedhi\Tools\EntityGenerator;

use Doctrine\Common\Inflector\Inflector as DoctrineInflector;

use Zeedhi\Libraries\ClassGenerator\PHPObjectValue;
use Zeedhi\Libraries\ClassGenerator\PHPType;
use Zeedhi\Libraries\ClassGenerator\PHPClass;
use Zeedhi\Libraries\ClassGenerator\PHPNamespace;
use Zeedhi\Libraries\ClassGenerator\PHPProperty;
use Zeedhi\Libraries\ClassGenerator\PHPBlock;
use Zeedhi\Libraries\ClassGenerator\PHPMethod;
use Zeedhi\Libraries\ClassGenerator\PHPParameter;
use Zeedhi\Libraries\ClassGenerator\PHPValue;
use Zeedhi\Libraries\ClassGenerator\Type;
use Zeedhi\Tools\Holder;
use Zeedhi\Tools\Nominator;
use Zeedhi\Tools\Generator\GeneratorConfig;

class ClassGenerator {

    const OVERWRITE_FILE = true;
    const KEEP_FILE = false;

    /** @var string */
    protected $psr4Root;

    public function __construct($verbose) {
        $this->verbose = $verbose;
    }

    public function setPsr4Root($psr4Root) {
        $this->psr4Root = $psr4Root;
    }

    protected function log($message, array $context = array()) {
        if($this->verbose){
            echo "[log] ", str_replace(array_keys($context), $context, $message), PHP_EOL;
        }
    }

    /**
     * @param string   $destinationFolder
     * @param PHPClass $class
     * @param boolean  $overWriteFile
     */
    protected function printToFile($destinationFolder, $class, $overWriteFile = self::KEEP_FILE) {
        $psr4 = $this->psr4Root."\\";
        $namespace = preg_replace('/^' . preg_quote($psr4, '/') . '/', '', $class->getNamespace()->fullName());
        $dir = $destinationFolder."\\".$namespace;
        $className = $class->getName();

        $dir = implode(DIRECTORY_SEPARATOR, explode("\\", $dir));
        $file = $dir.DIRECTORY_SEPARATOR.$className.".php";

        if (!is_dir($dir)) {
            mkdir($dir, 0777, true);
        }

        if ($overWriteFile || !file_exists($file)) {
            file_put_contents($file, $class->asPHP());
        }
    }


    protected function getRepositoryDummyClass() {
        $dummy = new PHPClass("EntityRepository");
        $dummy->setNamespace(new PHPNamespace("ORM", new PHPNamespace("Doctrine")));
        return $dummy;
    }

    protected function createTopClass(Holder\Entity $entity, $destinationFolder, $baseClass = null) {
        $baseClass = $baseClass ?: $this->createBaseClass($entity, $destinationFolder);
        $class = $entity->getClass();
        $phpClass = new PHPClass($class->getName());
        $phpClass->setNamespace($this->createPHPNamespaceFromZhNamespace($class->getNamespace()));
        $phpClass->setParentClass($baseClass);
        $this->printToFile($destinationFolder, $phpClass);
        $this->log("Top class created for entity {entity}", array("entity" => $entity->getName()));
    }

    protected function factoryPHPTypeFromEntityType($type) {
        static $PHPTypeFromEntityType = array(
            Holder\EntityField::TYPE_BLOB    => Type\Primary::TYPE_STRING,
            Holder\EntityField::TYPE_TEXT    => Type\Primary::TYPE_STRING,
            Holder\EntityField::TYPE_BOOLEAN => Type\Primary::TYPE_BOOLEAN,
            Holder\EntityField::TYPE_FLOAT   => Type\Primary::TYPE_FLOAT,
            Holder\EntityField::TYPE_INTEGER => Type\Primary::TYPE_INTEGER,
            Holder\EntityField::TYPE_STRING  => Type\Primary::TYPE_STRING
        );

        if ($type === Holder\EntityField::TYPE_DATETIME || $type === Holder\EntityField::TYPE_DATE) {
            $PHPType = new Type\Object("\\Datetime");
        } else {
            $PHPType = new Type\Primary($PHPTypeFromEntityType[$type]);
        }

        return $PHPType;
    }

    protected function createPHPNamespaceFromZhNamespace(Nominator\ZhNamespace $namespace) {
        $parentNamespace = $namespace->getParentNamespace();
        $parentPHPNamespace = $parentNamespace ? $this->createPHPNamespaceFromZhNamespace($parentNamespace): null;
        return new PHPNamespace($namespace->getName(), $parentPHPNamespace);
    }

    /**
     * @param PHPClass       $phpClass
     * @param string         $methodName
     * @param PHPParameter[] $parameters
     * @param string         $bodyCode
     */
    protected function addMethod(PHPClass $phpClass, $methodName, $parameters, $bodyCode) {
        $body = new PHPBlock($bodyCode);
        $method = new PHPMethod($methodName, $body, $parameters);
        $phpClass->addMethod($method);
    }

    /**
     * @param PHPClass $phpClass
     * @param $propertyName
     */
    protected function addGetMethod(PHPClass $phpClass, $propertyName) {
        $this->addMethod(
            $phpClass,
            "get" . ucfirst($propertyName),
            array(),
            'return $this->' . $propertyName . ';'
        );
    }

    /**
     * @param PHPClass $phpClass
     * @param string   $propertyName
     * @param PHPType  $PHPType
     */
    protected function addAddMethod(PHPClass $phpClass, $propertyName, $PHPType) {
        $singularPropertyName = DoctrineInflector::singularize($propertyName);
        $this->addMethod(
            $phpClass,
            "add" . ucfirst($singularPropertyName),
            array(new PHPParameter($singularPropertyName, new Type\Object($PHPType->getName()))),
            '$this->' . $propertyName . '[] = $' . $singularPropertyName . ';'
        );
    }

    /**
     * @param PHPClass $phpClass
     * @param $propertyName
     * @param $PHPType
     * @param $isNullAble
     */
    protected function addSetMethod(PHPClass $phpClass, $propertyName, $PHPType, $isNullAble) {
        $nullAblePHPValue = $isNullAble ? new PHPValue(null) : null;
        $this->addMethod(
            $phpClass,
            "set" . ucfirst($propertyName),
            array(new PHPParameter($propertyName, $PHPType, $nullAblePHPValue)),
            '$this->' . $propertyName . ' = $' . $propertyName . ';'
        );
    }

    /**
     * @param PHPClass $phpClass
     * @param string   $propertyName
     * @param PHPType  $PHPType
     * @param PHPValue $defaultValue
     * @param boolean  $isNullAble
     */
    protected function addPropertyToPHPClass(PHPClass $phpClass, $propertyName, $PHPType, $defaultValue = null, $isNullAble = false) {
        $phpClass->addProperty(new PHPProperty($propertyName, $PHPType, PHPProperty::VISIBILITY_PROTECTED, $defaultValue));
        $this->addGetMethod($phpClass, $propertyName);
        if ($PHPType->isCollection()) {
            $this->addAddMethod($phpClass, $propertyName, $PHPType);
        } else {
            $this->addSetMethod($phpClass, $propertyName, $PHPType, $isNullAble);
        }
    }

    protected function factoryPropertyForEntityRelation(Holder\EntityRelation $entityRelation, PHPClass $phpClass) {
        $isCollection = $entityRelation->getIsInversed() === 'Y';
        $relatedEntity = $entityRelation->getRelatedEntity();
        if($relatedEntity){
            $targetClassName = "\\" . $relatedEntity->getClass()->getFullName();
            $PHPType = new Type\Object($targetClassName, $isCollection);
            $propertyName = $entityRelation->getName();
            $isNullAble = $entityRelation->getEntityField()->getIsNullable() === 'Y';
            $defaultValue = $isCollection ? new PHPValue(array()) : null;
            $this->addPropertyToPHPClass($phpClass, $propertyName, $PHPType, $defaultValue, $isNullAble);
        }
    }

    /**
     * @param Holder\EntityField $entityField
     *
     * @return float|int|string
     */
    protected function parsePrimaryTypeValue(Holder\EntityField $entityField) {
        $value = $entityField->getDefaultValue();
        switch($entityField->getType()) {
            case Holder\EntityField::TYPE_STRING:
                $value = trim($value, "'");
                break;
            case Holder\EntityField::TYPE_INTEGER:
                $value = intval($value);
                break;
            case Holder\EntityField::TYPE_FLOAT:
                $value = floatval($value);
                break;
        }

        return $value;
    }

    /**
     * @param Holder\EntityField $entityField
     * @return PHPValue|null
     */
    protected function factoryDefaultValueForProperty(Holder\EntityField $entityField) {
        if ($entityField->getIsFk() === 'Y') return null;
        switch($entityField->getDefaultValue()) {
            case Holder\EntityField::DEFAULT_VALUE_CURRENT_DATE:
                $defaultValue = new PHPObjectValue("\\DateTime");
                break;
            case Holder\EntityField::DEFAULT_VALUE_NULL:
            case null:
                $defaultValue =  null;
                break;
            default:
                $defaultValue = new PHPValue($this->parsePrimaryTypeValue($entityField));
                break;
        }

        return $defaultValue;
    }

    /**
     * @param Holder\EntityField $entityField
     * @param PHPClass      $phpClass
     *
     * @return string a code to be added in constructor body.
     */
    protected function factoryPropertyForEntityField(Holder\EntityField $entityField, PHPClass $phpClass) {
        $constructorCode = "";
        $PHPType = $this->factoryPHPTypeFromEntityType($entityField->getType());
        $propertyName = $entityField->getName();
        $defaultValue = $this->factoryDefaultValueForProperty($entityField);
        if ($defaultValue instanceof PHPObjectValue) {
            //should be added to constructor body.
            $constructorCode = "\$this->{$propertyName} = " . $defaultValue->asPHP() . ";";
            $defaultValue = null;
        }

        $this->addPropertyToPHPClass($phpClass, $propertyName, $PHPType, $defaultValue, $entityField->getIsNullable() === 'Y');
        return $constructorCode;
    }

    /**
     * @param Holder\Entity $entity
     * @param PHPClass $phpClass
     */
    protected function factoryRelationsProperties(Holder\Entity $entity, PHPClass $phpClass) {
        $table = $entity->getTable();
        $tableName = $table->getName();

        /** @var Holder\EntityRelation[] $entityRelations */
        $entityRelations = isset($this->metaData["ENTITYRELATION"][$tableName]) ?
            $this->metaData["ENTITYRELATION"][$tableName] :
            [];

        foreach ($entityRelations as $constraintName => $relationColumns) {
            foreach ($relationColumns as $columnsName => $entityRelation) {
                $this->factoryPropertyForEntityRelation($entityRelation, $phpClass);
            }
        }
    }

    /**
     * @param Holder\Entity $entity
     * @param PHPClass $phpClass
     */
    protected function factoryFieldProperties(Holder\Entity $entity, PHPClass $phpClass) {
        $constructorCode = "";
        $table = $entity->getTable();
        $tableName = $table->getName();
        /** @var Holder\EntityField[] $entityFields */
        $entityFields = $this->metaData["FIELD"][$tableName];
        foreach ($entityFields as $fieldName=>$entityField) {
            // Fk Fields are created latter in relations.
            if ($entityField->getIsFk() === 'Y' && $entityField->getIsId() === 'N') continue;
            $propertyConstructorCode = $this->factoryPropertyForEntityField($entityField, $phpClass);
            if ($propertyConstructorCode) {
                $constructorCode .= $propertyConstructorCode . PHP_EOL;
            }
        }

        if ($constructorCode) {
            $phpClass->addMethod(new PHPMethod("__construct", new PHPBlock(trim($constructorCode, PHP_EOL))));
        }
    }

    /**
     * @param Holder\Entity $entity
     * @param PHPClass $phpClass
     */
    protected function factoryProperties(Holder\Entity $entity, PHPClass $phpClass) {
        $this->factoryFieldProperties($entity, $phpClass);
        $this->factoryRelationsProperties($entity, $phpClass);
    }

    /**
     * @param PHPClass[] $classesToImplement
     * @param PHPCLass   $phpClass
     */
    protected function addInterfacesToBaseClass($classesToImplement, PHPClass $phpClass) {
        foreach ($classesToImplement as $classToImplement) {
            $phpClass->addImplementedClass($classToImplement);
        }
    }

    protected function createBaseClass(Holder\Entity $entity, $destinationFolder, $classesToImplement = array()) {
        $class = $entity->getClass();
        $phpClass = new PHPClass($class->getName());
        $phpClass->setIsAbstract(true);
        $phpClass->setNamespace(new PHPNamespace("Base", $this->createPHPNamespaceFromZhNamespace($class->getNamespace())));
        $this->addInterfacesToBaseClass($classesToImplement, $phpClass);
        $this->factoryProperties($entity, $phpClass);
        $this->printToFile($destinationFolder, $phpClass, self::OVERWRITE_FILE);
        $this->log("Base class created for entity {entity}", array("entity" => $entity->getName()));
        return $phpClass;
    }

    protected function createRepositoryClass(Holder\Entity $entity, $destinationFolder) {
        $class = $entity->getRepositoryClass();
        $phpClass = new PHPClass($class->getName());
        $phpClass->setNamespace($this->createPHPNamespaceFromZhNamespace($class->getNamespace()));
        $phpClass->setParentClass($this->getRepositoryDummyClass());
        $this->printToFile($destinationFolder, $phpClass);
        $this->log("Repository class created for entity {entity}", array("entity" => $entity->getName()));
    }

    protected function getListenerBaseClass() {
        return new PHPClass('EntityListenerBaseClass');
    }

    protected function createEntityListenerClass(Holder\Entity $entity, $destinationFolder) {
        $class = $entity->getListenerClass();
        $phpClass = new PHPClass($class->getName());
        $phpClass->setNamespace($this->createPHPNamespaceFromZhNamespace($class->getNamespace()));
        $phpClass->setParentClass($this->getListenerBaseClass());
        $this->printToFile($destinationFolder, $phpClass, self::OVERWRITE_FILE);
        $this->log("Listener class created for entity {entity}", array("entity" => $entity->getName()));
        return array();
    }

    protected function createClassFiles($entity, $destinationFolder) {
        $this->log("Started to created entity {entity} files.", array("entity" => $entity->getName()));
        $classesToImplement = $this->createEntityListenerClass($entity, $destinationFolder);
        $baseClass = $this->createBaseClass($entity, $destinationFolder, $classesToImplement);
        $this->createTopClass($entity, $destinationFolder, $baseClass);
        $this->createRepositoryClass($entity, $destinationFolder);
        $this->log("Entity {entity} files created.", array("entity" => $entity->getName()));
    }

    /**
     * Generates mapping (xml, yaml or php) for entities.
     *
     * @param ConnectionConfig $config        The database and mapping configuration.
     * @param mixed[]          $metaData      The entities, fields, entity fields and entitiy relations objects.
     * @param mixed[]          $entitiesToUse The entities that must be generated.
     */
    public function createClassesFiles($config, $metaData, $entitiesToUse) {
        $nrEntitiesCreated = 0;
        $this->log("Started to create entities files.");

        $this->metaData = $metaData;
        $destinationFolder = $config->getSrcPath();

        foreach ($entitiesToUse as $entityName => $entity) {
            $this->createClassFiles($entity, $destinationFolder);
            $this->log("{entityName} entity created.", array("entityName" => $entityName));
        }
    }
}